/*
 * lab5.c
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>

#include "spi.h"
#include "ST7735R_TFT.h"
#include "usart.h"
#include "DHT.h"

static double humidity[1];
static double temperature[1];
static unsigned long last_screen_update = 0;

/* definitii note (Hz) */
#define c    261
#define d    294
#define e    329
#define f    349
#define g    391
#define gS   415
#define a    440
#define aS   455
#define b    466
#define cH   523
#define cSH  554
#define dH   587
#define dSH  622
#define eH   659
#define fH   698
#define fSH  740
#define gH   784
#define gSH  830
#define aH   880

uint16_t surprise_notes[] = {a, a, a, f, cH, a, f, cH, a, eH, eH, eH, fH, cH, gS, f, cH, a, aH, a, a, aH, gSH, gH, fSH, fH, fSH, aS, dSH, dH, cSH, cH, b, cH, f, gS, f, a, cH, a, cH, eH, aH, a, a, aH, gSH, gH, fSH, fH, fSH, aS, dSH, dH, cSH, cH, b, cH, f, gS, f, cH, a, f, cH, a};

uint16_t durations[] = {500, 500, 500, 350, 150, 500, 350, 150, 650, 500, 500, 500, 350, 150, 500, 350, 150, 650, 500, 300, 150, 400, 200, 200, 125, 125, 250, 250, 400, 200, 200, 125, 125, 250, 125, 500, 375, 125, 500, 375, 125, 650, 500, 300, 150, 400, 200, 200, 125, 125, 250, 250, 400, 200, 200, 125, 125, 250, 250, 500, 375, 125, 500, 375, 125, 650};

uint8_t num_notes = sizeof(durations) / sizeof(durations[0]);

volatile unsigned long milliseconds = 0;
volatile unsigned long next_alarm_time_left = 0; // in seconds
volatile uint8_t should_alarm_ring = 0;
volatile uint8_t has_not_yet_started = 0;

ISR(TIMER0_COMPA_vect)
{
    ++milliseconds;
    if (next_alarm_time_left != 0 && milliseconds % 1000 == 0) {
        next_alarm_time_left -= 1;
        if (next_alarm_time_left == 0) {
            should_alarm_ring = 1;
            has_not_yet_started = 1;
        }
    }
}

unsigned long millis(void)
{
    return milliseconds;
}

void IO_init()
{
    /* speaker-ul este puse ca iesire */
    DDRD |= (1 << PD4);

    /* butoanele sunt puse ca intrare, cu rezistenta de pull-up */
    DDRB &= ~(1 << PB2);
    PORTB |= (1 << PB2);
}

void timer0_init() {
    TCCR0B = (3 << CS00); // 64 prescaler
    TCCR0A = (1 << WGM01); // CTC
    OCR0A = 0xFA; // 250
    TIMSK0 = (1 << OCIE0A);
}

void timer1_init()
{
    TCCR1B |= (1 << CS10);

    TCCR1A |= (1 << COM1A1) | (1 << COM1A0);

    ICR1 = 0xFFFF;

    TCCR1A |= (1 << COM1B1) | (1 << WGM10) | (1 << WGM11);
    TCCR1B |= (1 << WGM12) | (1 << WGM13) | (1 << CS12);
}

int nr_nota;
int durata = 0;
void update_song()
{
    durata -= 25;
    if (durata < 25) {
        OCR1A = 15625 / surprise_notes[nr_nota];
        OCR1B = OCR1A / 2;
        TCNT1 = 0;
        durata = durations[nr_nota];
        nr_nota = (nr_nota + 1) % 64;
    }
}

void screen_update(void)
{
    static const char spaces[] = "                     ";
    uint8_t line_y = 20;
    char buf[32];

    sprintf(buf, "Temp is: %.2lf C", temperature[0]);
    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 255, 0, 0, 0, 0);
    ST7735R_DrawText(0, line_y, buf, 0, 255, 0, 0, 0, 0);

    sprintf(buf, "Hum is: %.2lf %%", humidity[0]);
    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 255, 0, 0, 0, 0);
    ST7735R_DrawText(0, line_y, buf, 0, 255, 0, 0, 0, 0);

    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    ST7735R_DrawText(0, line_y, "Time since start", 0, 0, 255, 0, 0, 0);

    unsigned long current_time = millis() / 1000;
    sprintf(buf, "%lu min %lu sec", current_time / 60, current_time % 60);
    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    ST7735R_DrawText(0, line_y, buf, 0, 0, 255, 0, 0, 0);

    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    line_y += ST7735R_FONT_HEIGHT;
    ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    if (next_alarm_time_left == 0) {
        ST7735R_DrawText(0, line_y, "No alarm set", 255, 0, 0, 0, 0, 0);
        line_y += ST7735R_FONT_HEIGHT;
        ST7735R_DrawText(0, line_y, spaces, 0, 0, 255, 0, 0, 0);
    } else {
        ST7735R_DrawText(0, line_y, "Next alarm in", 255, 0, 0, 0, 0, 0);

        sprintf(buf, "%lu min %lu sec", next_alarm_time_left / 60, next_alarm_time_left % 60);
        line_y += ST7735R_FONT_HEIGHT;
        ST7735R_DrawText(0, line_y, spaces, 255, 0, 0, 0, 0, 0);
        ST7735R_DrawText(0, line_y, buf, 255, 0, 0, 0, 0, 0);
    }
}

void process_request(const char *request)
{
    uint16_t min, sec;
    sscanf(request, "a%um%usf", &min, &sec);
    next_alarm_time_left = min * 60 + sec;
}

int main(void)
{

    // Init SPI for LCD.
    SPI_init();
    USART0_init();
    DHT_Setup();

    // Init LCD
    ST7735R_Begin();

    // Clear screen with BLACK.
    ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 0, 0, 0);

    sei();

    IO_init();
    timer0_init();

    for(;;) {
        if (millis() - last_screen_update > 250) {
            DHT_Read(temperature, humidity);
            screen_update();
            last_screen_update = millis();
        }

        if (should_alarm_ring == 1) {
            if (has_not_yet_started == 1) {
                timer1_init();
                OCR1A = 0xFFFF;
                OCR2A = 0xFF;
                has_not_yet_started = 0;
            }
            update_song();
        }

        _delay_ms(25);

        if (!(PINB & (1 << PB2))) {
            char req[128];
            int i = 0;
            do {
                req[i] = USART0_receive();
                i += 1;
            } while(req[i - 1] != 'f' && i < 127);
            req[i] = '\0';
            
            process_request(req);
        }
    }

    /* A happy compiler is a healthy compiler. */
    return 0;
}
